/* 
 * This file is part of the ARM Cortex-M Fault distribution.
 * Copyright (c) 2018 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "mpu.h"

struct mpu {
	unsigned int TYPE;
	unsigned int CTRL;
	unsigned int RNR;
	unsigned int RBAR;
	unsigned int RASR;
	unsigned int RBAR_A1;
	unsigned int RASR_A1;
	unsigned int RBAR_A2;
	unsigned int RASR_A2;
	unsigned int RBAR_A3;
	unsigned int RASR_A3;
};

enum reg_ctrl {
	CTRL_PRIVDEFENA = 1 << 2,
	CTRL_HFNMIENA = 1 << 1,
	CTRL_ENABLE = 1 << 0,
};

enum reg_rasr {
	RASR_XN = 1 << 28,
	RASR_AP_SHIFT = 24,
	RASR_TEXSCB_SHIFT = 16,
	RASR_SRD_SHIFT = 8,
	RASR_SIZE_SHIFT = 1,
};

static volatile struct mpu * const mpu = (struct mpu *)0xE000ED90;

void mpu_onoff(unsigned int onoff){
	
	//Membarrier
	asm("DSB");
	asm("ISB");
	
	unsigned int reg = mpu->CTRL;
	reg &= ~CTRL_ENABLE;
	reg |= CTRL_PRIVDEFENA;
	mpu->CTRL |= onoff & 0b1;
	
	//Membarrier
	asm("DSB");
	asm("ISB");
}


void mpu_set_region(	
		unsigned int region, 
		unsigned int baseaddr, 					//startet bei vielfachen von size
		enum mpu_regionsize regionsize_bits,	//regionsize_bits -> bytes in 2^(size+1)
		unsigned int subregionbits,				//Bitmaske von 8 Bits, erst nutzbar ab mehr als 128 Byte Regionsgre
		enum mpu_regionattributes regionattributes, 
		enum mpu_permissions permissions,
		unsigned int executable){
		
	//Membarrier
	asm("DSB");
	asm("ISB");

	//angeben welche Region eingestellt werden soll
	mpu->RNR = region & 0x7;
	
	//Region abschalten vor nderung!
	mpu->RASR = 0;
	
	//Basisadresse reinwerfen
	mpu->RBAR = baseaddr & 0xFFFFFFE0;
	
	unsigned int reg_rasr = 1; //Enable Bit
	
	if (0 == executable){
		reg_rasr |= RASR_XN;
	}
	
	reg_rasr |= (permissions & 0b111) << RASR_AP_SHIFT;
	
	reg_rasr |= (regionattributes & 0b111111) << RASR_TEXSCB_SHIFT;
	
	reg_rasr |= (subregionbits & 0xFF) << RASR_SRD_SHIFT;
	
	reg_rasr |= (regionsize_bits & 0b11111) << RASR_SIZE_SHIFT;
	
	mpu->RASR = reg_rasr;
	
	//Membarrier
	asm("DSB");
	asm("ISB");
}