/* 
 * This file is part of the ARM Cortex-M Fault distribution.
 * Copyright (c) 2018 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdint.h>

#include "scb.h"

#include "armcore/nvic.h"
#include "lib/deprintf.h"

struct scb {
	unsigned int CPUID;	// CPUID Base Register
	unsigned int ICSR;	// Interrupt Control and State Register
	unsigned int VTOR;	// Vector Table Offset Register
	unsigned int AIRCR;	// Application Interrupt and Reset Control Register
	unsigned int SCR;	// Application Interrupt and Reset Control Register
	unsigned int CCR;	// Configuration and Control Register
	unsigned int SHPR1;	// System Handler Priority Register 1
	unsigned int SHPR2;	// System Handler Priority Register 2
	unsigned int SHPR3;	// System Handler Priority Register 3
	unsigned int SHCRS;	// System Handler Control and State Register
	unsigned int CFSR;	// besteht aus: UFSR, BFSR, MMSR
	unsigned int HFSR;	// HardFault Status Register
	unsigned int unused;
	unsigned int MMAR;	// MemManage Fault Address Register
	unsigned int BFAR;	// BusFault Address Register
	unsigned int AFSR;	// Auxiliary Fault Status Register
};

static volatile struct scb * const scb = (struct scb *)0xE000ED00;

enum reg_ccr {
	CCR_DIV_0_TRP 	= 1 << 4,
	CCR_UNALIGN_TRP = 1 << 3,
};

enum reg_shcrs {
	SHCRS_USGFAULTENA 		= 1 << 18,
	SHCRS_BUSFAULTENA 		= 1 << 17,
	SHCRS_MEMFAULTENA 		= 1 << 16,
	SHCRS_SVCALLPENDED 		= 1 << 15,
	SHCRS_BUSFAULTPENDED 	= 1 << 14,
	SHCRS_MEMFAULTPENDED 	= 1 << 13,
	SHCRS_USGFAULTPENDED 	= 1 << 12,
	SHCRS_SYSTICKACT 		= 1 << 11,
	SHCRS_PENDSVACT 		= 1 << 10,
	SHCRS_MONITORACT 		= 1 << 8,
	SHCRS_SVCALLACT 		= 1 << 7,
	SHCRS_USGFAULTACT 		= 1 << 3,
	SHCRS_BUSFAULTACT 		= 1 << 1,
	SHCRS_MEMFAULTACT 		= 1 << 0,
};

enum reg_ufsr {
	UFSR_DIVBYZERO 	= 1 << 9,	// the processor has executed an SDIV or UDIV instruction with a divisor of 0
	UFSR_UNALIGNED 	= 1 << 8,	// the processor has made an unaligned memory access
	UFSR_NOCP 		= 1 << 3,	// the processor has attempted to access a coprocessor
	UFSR_INVPC 		= 1 << 2,	// the processor has attempted an illegal load of EXC_RETURN to the PC
	UFSR_INVSTATE 	= 1 << 1,	// the processor has attempted to execute an instruction that makes illegal use of the EPSR
	UFSR_UNDEFINSTR = 1 << 0,	// the processor has attempted to execute an undefined instruction
};

enum reg_bfsr {
	BFSR_BFARVALID 	= 1 << 7,	// BFAR holds a valid fault address
	BFSR_STKERR 	= 1 << 4,	// stacking for an exception entry has caused one or more BusFaults
	BFSR_UNSTKERR 	= 1 << 3,	// unstack for an exception return has caused one or more BusFaults
	BFSR_IMPRECISERR = 1 << 2,	// a data bus error has occurred, but the return address in the stack frame is not related to the instruction that caused the error
	BFSR_PRECISERR 	= 1 << 1,	// a data bus error has occurred, and the PC value stacked for the exception return points to the instruction that caused the fault
	BFSR_IBUSERR 	= 1 << 0,	// instruction bus error
};

enum reg_mmfsr {
	MMSR_MMARVALID 	= 1 << 7,	// MMAR holds a valid fault address
	MMSR_MSTKERR 	= 1 << 4,	// stacking for an exception entry has caused one or more access violations
	MMSR_MUNSTKERR 	= 1 << 3,	// unstack for an exception return has caused one or more access violations
	MMSR_DACCVIOL 	= 1 << 1,	// the processor attempted a load or store at a location that does not permit the operation
	MMSR_IACCVIOL 	= 1 << 0,	// the processor attempted an instruction fetch from a location that does not permit execution
};

enum reg_hfsr {
	HFSR_FORCED_SHIFT 	= 30,	// Hardfault was forced
	HFSR_FORCED 		= 1 << 30,
};

// the prioritys of the System Handlers starting with 4
#define SHPR_PRIO_START 4

#define UFSR_MSK 0xFFFF0000
#define UFSR_POS 16
#define BFSR_MSK 0x0000FF00
#define BFSR_POS 8
#define MMSR_MSK 0x000000FF
#define MMSR_POS 0

void scb_en_div0_trap(void){
	scb->CCR |= CCR_DIV_0_TRP;
}

void scb_en_unaligned_trap(void){
	scb->CCR |= CCR_UNALIGN_TRP;
}

void scb_vtor(unsigned int addr, enum scb_vtor_place place){
	scb->VTOR = addr | place;
	asm("DSB");
}

void scb_enable_faulthandler(enum fhandler handle, unsigned int preempt_prio, unsigned int sub_prio){

	// Zuerst Prio setzen
	uint8_t prio = preempt_prio << NVICIRQ_PREEMPTBITS;
	prio |= sub_prio & NVICIRQ_SUBPRIOMASK;
	
	// Arrayindex raussuchen fr SHPRx (startet mit Prio 4 als erstes)
	volatile uint8_t *array = (uint8_t*)&(scb->SHPR1);
	array[handle - SHPR_PRIO_START] = prio;
	
	unsigned int setbit = 0;
	
	switch (handle){
		case HANDLER_MEMMANAGE:
			setbit = SHCRS_MEMFAULTENA;
			break;
		case HANDLER_BUSFAULT:
			setbit = SHCRS_BUSFAULTENA;
			break;
		case HANDLER_USAGEFAULT:
			setbit = SHCRS_USGFAULTENA;
			break;
		case HANDLER_SVCCALL:
		case HANDLER_PENDSV:
		case HANDLER_SYSTICK:
			// ist immer an! -> Nur prio setzen
			return;
	}
	
	scb->SHCRS |= setbit;
}

unsigned int scb_hfault_was_forced(void){

	return (scb->HFSR & HFSR_FORCED) >> HFSR_FORCED_SHIFT;
}

void scb_print_ufsr_faults(void){
	
	// UFSR ist ein Teil von CFSR
	unsigned int ufsr = (scb->CFSR & UFSR_MSK) >> UFSR_POS;
	
	if (ufsr & UFSR_DIVBYZERO){
		deprintf("DIVBYZERO ");
	}
	
	if (ufsr & UFSR_UNALIGNED){
		deprintf("UNALIGNED ");
	}
	
	if (ufsr & UFSR_NOCP){
		deprintf("NO_COPROC ");
	}
	
	if (ufsr & UFSR_INVPC){
		deprintf("INVALID_PC ");
	}
	
	if (ufsr & UFSR_INVSTATE){
		deprintf("INVSTATE ");
	}
	
	if (ufsr & UFSR_UNDEFINSTR){
		deprintf("UNDEFINSTR ");
	}
}

void scb_print_bfsr_faults(void){
	
	// BFSR ist ein Teil von CFSR
	unsigned int bfsr = (scb->CFSR & BFSR_MSK) >> BFSR_POS;
	
	if (bfsr & BFSR_STKERR){
		deprintf("STKERR ");
	}
	
	if (bfsr & BFSR_UNSTKERR){
		deprintf("UNSTKERR ");
	}
	
	if (bfsr & BFSR_IMPRECISERR){
		deprintf("IMPRECISERR ");
	}
	
	if (bfsr & BFSR_PRECISERR){
		deprintf("PRECISERR ");
	}
	
	if (bfsr & BFSR_IBUSERR){
		deprintf("IBUSERR ");
	}
}

void scb_print_mmsr_faults(void){
	
	// MMSR ist ein Teil von CFSR
	unsigned int mmsr = (scb->CFSR & MMSR_MSK) >> MMSR_POS;
	
	if (mmsr & MMSR_MSTKERR){
		deprintf("MSTKERR ");
	}
	
	if (mmsr & MMSR_MUNSTKERR){
		deprintf("MUNSTKERR ");
	}
	
	if (mmsr & MMSR_DACCVIOL){
		deprintf("DACCVIOL ");
	}
	
	if (mmsr & MMSR_IACCVIOL){
		deprintf("IACCVIOL ");
	}
}

void scb_print_cfsr_faults(void){

	// CFSR in seine Subregister aufsplitten
	unsigned int ufsr = (scb->CFSR & UFSR_MSK) >> UFSR_POS;
	unsigned int bfsr = (scb->CFSR & BFSR_MSK) >> BFSR_POS;
	unsigned int mmsr = (scb->CFSR & MMSR_MSK) >> MMSR_POS;
	
	// wenn in den SUbrigstern bits gesetzt sind, dann ausgeben
	
	if (ufsr){
		deprintf("\nUsageFault: ");
		scb_print_ufsr_faults();
	}
	
	if (bfsr){
		deprintf("\nBusFault: ");
		scb_print_bfsr_faults();
	}
	
	if (mmsr){
		deprintf("\nMemManage Fault: ");
		scb_print_mmsr_faults();
	}

}

void scb_print_busfaultaddr(void){

	unsigned int bfsr = (scb->CFSR & BFSR_MSK) >> BFSR_POS;
	
	if (bfsr & BFSR_BFARVALID){
		deprintf("BusFault Address: %X\n", scb->BFAR);
	}
}

void scb_print_mmfaultaddr(void){

	unsigned int mmsr = (scb->CFSR & MMSR_MSK) >> MMSR_POS;
	
	if (mmsr & MMSR_MMARVALID){
		deprintf("MemManage Fault Address: %X\n", scb->MMAR);
	}
}