/* 
 * This file is part of the ARM Cortex-M Fault distribution.
 * Copyright (c) 2018 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdint.h>

// Trace Port Interface Unit
struct tpiu {
  unsigned int SSPSR;		// Supported Synchronous Port Size Register
  unsigned int CSPSR;		// Current Synchronous Port Size Register
  unsigned int unused0[2];
  unsigned int ACPR; 		// Asynchronous Clock Prescaler Register
  unsigned int unused1[55];	
  unsigned int SPPR;		// Selected Pin Protocol Register
  unsigned int unused2[131];
  unsigned int FFSR;		// Formatter and Flush Status Register
  unsigned int FFCR;		// Formatter and Flush Control Register
  unsigned int FSCR;		// Formatter Synchronization Counter Register
};

enum TPIU_SPPR_VALUES {
	SPPR_PARATRACE 		= 0,	// Parallel trace port mode.
	SPPR_SWO_MANCHESTER = 1,	// Asynchronous SWO, using Manchester encoding
	SPPR_SWO_NRZ		= 2,	// Asynchronous SWO, using NRZ encoding
};

struct itm {
  union {
	uint8_t u8;		// 8Bit schreiben fr die Zeichen FIFO
	uint32_t u32;	// 32Bit auslesen fr das Statusbit!
  } STIM[32];
  unsigned int unused0[864];
  unsigned int TER;	// Trace Enable Register
  unsigned int unused1[15];
  unsigned int TPR;	// Refer to the ITM Trace Privilege Register description
  unsigned int unused2[15];
  unsigned int TCR;	// Trace Control Register
  unsigned int unused3[29];
  unsigned int IWR;
  unsigned int IRR;
  unsigned int IMCR;
  unsigned int unused4[43];
  unsigned int LAR;	// Lock Access
  unsigned int LSR;	// Lock Status
};

#define ITM_PASSWORD 0xC5ACCE55

struct coredebug {
  unsigned int DFSR;
  unsigned int DHCSR;
  unsigned int DCRSR;
  unsigned int DCRDR;
  unsigned int DEMCR;
};

#define TPR_PRIVMASK 0xF // Alle Stimulus Ports aktiv

enum ITM_BITS {
  ITM_ATBID 		= 0x7F,
  ITM_ATBID_SHIFT 	= 16,
  ITM_SWOENA 		= 1 << 4,
  ITM_DWTENA 		= 1 << 3,
  ITM_SYNCENA 		= 1 << 2,
  ITM_TSENA 		= 1 << 1,
  ITM_ITMENA 		= 1 << 0,
};

#define CoreDebug_DEMCR_TRCENA_Msk (1<<24)

static volatile struct tpiu * const TPIU = (struct tpiu *)0xE0040000;
static volatile struct itm * const ITM = (struct itm *)0xE0000000;
static volatile struct coredebug * const CoreDebug = (struct coredebug *)0xE000EDF0;

void swo_init(unsigned int swo_freq, unsigned int cpu_freq) {
  
  unsigned int swo_prescaler = (cpu_freq/swo_freq) - 1;
  
  // Trace berhaupt ersteinmal aktivieren
  CoreDebug->DEMCR = CoreDebug_DEMCR_TRCENA_Msk;
 
  // SWO Pin Protokoll
  TPIU->SPPR = SPPR_SWO_NRZ;
  
  // SWO Pin Baudrate
  TPIU->ACPR = swo_prescaler; 
  
  // ITM Kontrollzugriffe mit Passwort freischalten (sonst ist nur Stimulus Port Zugriff mglich)
  ITM->LAR = ITM_PASSWORD;
  
  // Tracecontrol: SWO aktivieren, ITM aktivieren
  ITM->TCR =  ITM_SWOENA | ITM_ITMENA;
  
  // Trace Privileg: jeder darf alles
  ITM->TPR = TPR_PRIVMASK;
}

void swo_manage_stimulus_port(int stimulus_port, int onoff){
  
  // ITM Kontrollzugriffe freischalten (sonst ist nur Stimulus Port Zugriff mglich)
  ITM->LAR = ITM_PASSWORD;
  
  if (onoff){
    ITM->TER |= 1 << stimulus_port;
  }else{
    ITM->TER &= ~(1 << stimulus_port);
  }
}

void swo_printchar(int stimulus_port, char c){
  
  // Es gibt nur 32 Stimulus Ports
  if (stimulus_port >= 32){
    return;
  }
  
  // Wenn Trace deaktiviert, dann passiert hier nichts auer eine Endlosschleife
  if (0 == (ITM->TCR & ITM_ITMENA)){
    return;
  }
  
  // Nur wenn das passendes Trace Enable Register Bit gesetzt ist funktioniert die Ausgabe.
  // Sonst Endlosschleife
  if (0 == (ITM->TER & (1 << stimulus_port))){
    return;
  }
  
  // Wenn die FIFO des Stimulus Port nicht voll ist dann schreiben, sonst warten
  while ( 0 == ITM->STIM[stimulus_port].u32);
  ITM->STIM[stimulus_port].u8 = c;
}

// write Syscall berschreiben damit printf Zeichen ber SWO ausgeben kann
int _write(int file, char *ptr, int len){
	
	(void)file;
	int i;
	for (i = 0; i < len; i++){
		swo_printchar(0, ptr[i]);
	}
	
	return len;
}