#include "usart.h"
#include "usart_regdefs.h"

#include <stdlib.h>
#include <stdint.h>

#if USART_USE_DMA
	#include "periph/dma.h"
#endif

#if (USART_USE_IRQ || USART_USE_DMA)
	#include "armcore/nvic.h"
#endif

//USART Basisadressen
static unsigned int usart_bases[7] = 
	{0, USART1_BASE_INT, USART2_BASE_INT, USART3_BASE_INT, UART4_BASE_INT, UART5_BASE_INT, USART6_BASE_INT};

//Pointer auf FIFO idx merken fr IRQs
#if USART_USE_IRQ
static struct usart_irq *irq_idx[7] = 
	{NULL, NULL, NULL, NULL, NULL, NULL, NULL};

//IRQ Nummern zum Pending zurcksetzen
static uint8_t usart_irq_nbrs[7] = 
	{0, IRQ_NBR_USART1, IRQ_NBR_USART2, IRQ_NBR_USART3, IRQ_NBR_UART4, IRQ_NBR_UART5, IRQ_NBR_USART6};
#endif

/*
1. Enable the USART by writing the UE bit in USART_CR1 register to 1.
2. Program the M bit in USART_CR1 to define the word length.
3. Program the number of stop bits in USART_CR2.
4. <DMA foo>
5. Select the desired baud rate using the USART_BRR register.
6. Set the TE bit in USART_CR1 to send an idle frame as first transmission
*/
#if USART_USE_IRQ
void usart_init_irq(struct usart_irq *idx, unsigned int usart_base, unsigned int baudrate, unsigned int bus_clk){

	volatile struct usart * const usart = (struct usart *)usart_bases[usart_base];
	irq_idx[usart_base] = idx;
	idx->usart_base_addr = usart_bases[usart_base];
	
	usart->CR1 = CR1_UE;
	usart->CR2 = 0;
	usart->CR3 = 0;	
	
	/*
		Baudteiler ist bus_clk/(16 * baudrate)
		Fraction hat 4 bits -> 16
		-> sparen wir uns die 16 und daher diese Formel:
	*/
	usart->BRR = bus_clk/baudrate;
	
	usart->CR1 |= CR1_RXNEIE | CR1_TE | CR1_RE;
	
	nvic_enable_irq(usart_irq_nbrs[usart_base], 0xE, 0xF);
};
#endif

void usart_rebaud(unsigned int usart_base, unsigned int baudrate, unsigned int bus_clk){

	volatile struct usart * const usart = (struct usart *)usart_bases[usart_base];
	usart->BRR = bus_clk/baudrate;
};

void usart_init_poll(unsigned int usart_base, unsigned int baudrate, unsigned int bus_clk){

	volatile struct usart * const usart = (struct usart *)usart_bases[usart_base];
	
	#if USART_USE_IRQ
	irq_idx[usart_base] = NULL;
	nvic_disable_irq(usart_irq_nbrs[usart_base]);
	#endif
	
	usart->CR1 = CR1_UE; 	// killt auch folgende IRQs: PEIE, TXEIE, TCIE, RXNEIE, IDLEIE
	usart->CR2 = 0;
	usart->CR3 = 0;			// killt auch DMA
	
	usart->BRR = bus_clk/baudrate;
	
	usart->CR1 |= CR1_TE; 
}

#if USART_USE_DMA
int usart_dmasend(
	unsigned int usart_base,				// Welcher USART
	enum dma_periph peri,					// Welcher DMA/Stream/Channel
	uint8_t *buf,							// Buf zum senden
	uint16_t buflen,						// Elemente im Buf
	void (*callb)(unsigned int, void*),	// Callback be auftretendem DMA IRQ
	void *callb_data						// Daten fr den DMA Callback
	){
	
	volatile struct usart * const usart = (struct usart *)usart_bases[usart_base];
	
	int ret = dma_periph(
		peri,						// Angabe der Peripherie
		(unsigned int)buf,			// BUF Mem
		(unsigned int)&(usart->DR),// REG Periph
		buflen,						// Anzahl der zu bertragenden Elemente
		DMA_SIZE_BYTE,				// Gre der zu bertragenden Daten
		0,							// mem wird circulr befllt = DMA geht nie zuende
		DMA_MEM_TO_PERIPH,			// in welche Richtung geht der DMA (mem to mem ist verboten)
		DMA_PRIO_LOW,				// Prioritt des DMA
		callb,						// Callback be auftretendem IRQ
		callb_data					// Daten fr den Callback
		);
		
	if (ret){
		return 1;
	}
	
	usart->CR3 |= CR3_DMAT;
	
	return 0;
}
#endif

#if USART_USE_IRQ
void usart_putc(unsigned int usart_base, char c){

	struct usart_irq *idx = irq_idx[usart_base];
	if (0 == idx){
		return;
	}
	
	fifo_put_block(&(idx->tx_fifo), c);
	volatile struct usart * const usart = (struct usart *)idx->usart_base_addr;
	usart->CR1 |= CR1_TXEIE;
};

void usart_puts(unsigned int usart_base, char *str){

	struct usart_irq *idx = irq_idx[usart_base];
	if (0 == idx){
		return;
	}
	
	volatile struct usart * const usart = (struct usart *)idx->usart_base_addr;
	
	while (*str){
		fifo_put_block(&(idx->tx_fifo), *str++);
		usart->CR1 |= CR1_TXEIE;
	}
}

char usart_getc(unsigned int usart_base){

	struct usart_irq *idx = irq_idx[usart_base];
	if (0 == idx){
		return 0;
	}
	
	return fifo_get_block(&(idx->rx_fifo));
}
#endif

void usart_poll_putc(unsigned int usart_base, char c){

	volatile struct usart * const usart = (struct usart *)usart_bases[usart_base];
	
	while (0 == (usart->SR & SR_TXE));
	usart->DR = c;
}

//wird vom IRQ ASM Trampolin aufgerufen mit passender USART Nummer
#if USART_USE_IRQ
void usart_irqhandler(unsigned int usart_base){

	struct usart_irq *idx = irq_idx[usart_base];
	if (0 == idx){
		return;
	}
	
	volatile struct usart * const usart = (struct usart *)usart_bases[usart_base];
	unsigned int status = usart->SR;
	
	if (status & SR_RXNE){
		fifo_put(&(idx->rx_fifo), usart->DR);
	}
	
	if ((status & SR_TXE) && (0 == fifo_empty(&(idx->tx_fifo)))){
		usart->DR = fifo_get(&(idx->tx_fifo));
		if (fifo_empty(&(idx->tx_fifo))){
			usart->CR1 &= ~CR1_TXEIE;
		}
	}
	
	nvic_pending_clear(usart_irq_nbrs[usart_base]);
}

void usart1_irq(void){
	usart_irqhandler(1);
}

void usart2_irq(void){
	usart_irqhandler(2);
}

void usart3_irq(void){
	usart_irqhandler(3);
}

void usart4_irq(void){
	usart_irqhandler(4);
}

void usart5_irq(void){
	usart_irqhandler(5);
}

void usart6_irq(void){
	usart_irqhandler(6);
}

#endif