/* 
 * This file is part of the ARM Cortex-M Fault distribution.
 * Copyright (c) 2018 Martin Wende.
 * 
 * This program is free software: you can redistribute it and/or modify  
 * it under the terms of the GNU General Public License as published by  
 * the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
*/

#include "main.h"

#include "periph/rcc.h"
#include "periph/usart.h"
#include "periph/gpio.h"

#include "lib/deprintf.h"

#include "armcore/scb.h"

#include "system/disasm.h"

struct faultregs {
	unsigned int R4;
	unsigned int R5;
	unsigned int R6;
	unsigned int R7;
	unsigned int R8;
	unsigned int R9;
	unsigned int R10;
	unsigned int R11;
	unsigned int SP;
	unsigned int R0;
	unsigned int R1;
	unsigned int R2;
	unsigned int R3;
	unsigned int R12;
	unsigned int LR;
	unsigned int PC;
	unsigned int PSR;	
};

static char *sys_isr_str[16] = {
	"Thread mode",
	"",
	"NMI",
	"HardFault",
	"MemManage",
	"BusFault",
	"BusFault",
	"",
	"",
	"",
	"",
	"SVCall",
	"DebugMonitor",
	"",
	"PendSV",
	"SysTick"
};

static void set_fault_usart(void){
	
	/*
	Im Fault sind keine IRQs verfgbar
	zB knnte ein Timer per IRQ sonst wieder Code ausfhren, der zu einen weiteren Fault fhrt
	Daher muss der UART hier auf Pollbetrieb umgestellt werden
	*/
	
	// USART3 poll Renit fr deprintf
	rcc_enable_clock(RCC_USART3);
	rcc_enable_clock(RCC_GPIOC);
	gpio_initpin(PORTC, 10, AF_PP, GPIO_HS, AF7_USART1_TO_3); //Altfunc TX Pin
	usart_init_poll(USART3_INDEX, UART_BAUDRATE, SYS_CLK);
	
	deprintf_setusart(USART3_INDEX, usart_poll_putc);
}

static void sys_notaus(void){

	// Hier das System in ein definierten Zustand bringen der niemanden Umbringt
	// z.B. den Strom abschalten bei einer E-Last

}

static void print_psr(unsigned int psr){
	
	if (psr & (1 << 31)){
		deprintf("N");
	} else {
		deprintf("_");
	}
	
	if (psr & (1 << 30)){
		deprintf("Z");
	} else {
		deprintf("_");
	}
	
	if (psr & (1 << 29)){
		deprintf("C");
	} else {
		deprintf("_");
	}
	
	if (psr & (1 << 28)){
		deprintf("V");
	} else {
		deprintf("_");
	}
	
	if (psr & (1 << 27)){
		deprintf("Q");
	} else {
		deprintf("_");
	}
	
	if (psr & (1 << 24)){
		deprintf("T");
	} else {
		deprintf("_");
	}
	
	unsigned int isr_nbr = psr & 0x1FF;
	// ausgelesener ISR handler > 16 -> IRQ, kein Faulthandler
	if (isr_nbr > 16){
		deprintf(" IRQ Pos: %u", isr_nbr-16);
	} else {
		deprintf(" Syshandler: %u %s", isr_nbr, sys_isr_str[isr_nbr]);
	}
	
	deprintf(" (%X)", psr);
}

static void print_regdump(struct faultregs *regs){

	deprintf("\nRegisterdump: \n");
	deprintf("xPSR: ");
	print_psr(regs->PSR);
	deprintf("\n");
	
	deprintf("R0: %X   R4: %X   R8:  %X  R12: %X\n", regs->R0, regs->R4, regs->R8, regs->R12);
	deprintf("R1: %X   R5: %X   R9:  %X  SP:  %X\n", regs->R1, regs->R5, regs->R9, regs->SP);
	deprintf("R2: %X   R6: %X   R10: %X  LR:  %X\n", regs->R2, regs->R6, regs->R10, regs->LR);
	deprintf("R3: %X   R7: %X   R11: %X  PC:  %X\n", regs->R3, regs->R7, regs->R11, regs->PC);
}

void hardfault_handler(struct faultregs *regs){

	sys_notaus();		// System in definierten Zustand bringen z.B. Strom aus, Motor aus
	set_fault_usart();	// UART auf Poll Betrieb umschalten, IRQs stehen nicht mehr zur Verfgung
	
	unsigned int isforced = scb_hfault_was_forced();
	char fcd[] = "\0forced ";
	char *strp;
	
	// forced oder nicht?
	strp = &fcd[1];
	if (!isforced){
		strp = &fcd[0];
	}
	
	deprintf("\n\n###################################################\n");
	deprintf("%sHardFault! An Adresse: %X\n\n", strp, regs->PC);
	if (isforced){
		deprintf("forced Error(s): ");
		scb_print_cfsr_faults();  // Faultregister durch Unterfunktion ausgeben
		deprintf("\n");
	}
	
	// beim forced knnte eine Busadresse gespeichert sein
	scb_print_busfaultaddr();
	scb_print_mmfaultaddr();
	
	print_regdump(regs);
	
	deprintf("\nSystem halted!");

	while (1);	// Anhalten, alles ist kaputt
};

void mpufault_handler(struct faultregs *regs){

	sys_notaus();
	set_fault_usart();
	
	deprintf("\n\n###################################################\n");
	deprintf("MPUFault! An Adresse: %X\n", regs->PC);
	
	deprintf("\nCause: ");
	scb_print_mmsr_faults();
	deprintf("\n");
	scb_print_mmfaultaddr();
	deprintf("\n");
	
	print_regdump(regs);
	
	deprintf("\nSystem halted!");

	while (1);
};

void busfault_handler(struct faultregs *regs){

	sys_notaus();
	set_fault_usart();
	
	deprintf("\n\n###################################################\n");
	deprintf("BusFault! An Adresse: %X\n", regs->PC);
	
	deprintf("\nCause: ");
	scb_print_bfsr_faults();
	deprintf("\n");
	scb_print_busfaultaddr();
	deprintf("\n");
	
	print_regdump(regs);
	
	deprintf("\nSystem halted!");

	while (1);
};

void usagefault_handler(struct faultregs *regs){

	sys_notaus();
	set_fault_usart();
	
	deprintf("\n\n###################################################\n");
	deprintf("UsageFault! An Adresse: %X\n", regs->PC);
	
	deprintf("\nCause: ");
	scb_print_ufsr_faults();	// Usagefaultregister durch Unterfunktion ausgeben
	deprintf("\n");
	
	print_regdump(regs);
	deprintf("\nvergeigter Befehl: ");
	disasm_print(regs->PC);
	
	deprintf("\nSystem halted!");

	while (1);
};